/* This is a Cfunctions (version 0.28) generated header file.
   Cfunctions is a free program for extracting headers from C files.
   Get Cfunctions from 'http://www.lemoda.net/cfunctions/'. */

/* This file was generated with:
'cfunctions -inc edit-distance-char.c' */
#ifndef CFH_EDIT_DISTANCE_CHAR_H
#define CFH_EDIT_DISTANCE_CHAR_H

/* From 'edit-distance-char.c': */

#line 11 "edit-distance-char.c"
int distance_char (const unsigned char * word1 , int len1 , text_fuzzy_t * tf );

#endif /* CFH_EDIT_DISTANCE_CHAR_H */
